'use strict'
var __importDefault =
	(this && this.__importDefault) ||
	function (mod) {
		return mod && mod.__esModule ? mod : { default: mod }
	}
Object.defineProperty(exports, '__esModule', { value: true })
// Import
const util_1 = __importDefault(require('util'))
const assert_1 = __importDefault(require('assert'))
const ansicolors_1 = __importDefault(require('ansicolors'))
const diff_1 = require('diff')
const errlop_1 = __importDefault(require('errlop'))
/** Alias for setTimeout with paramaters reversed. */
function wait(delay, fn) {
	return setTimeout(fn, delay)
}
exports.wait = wait
/** Whether or not we are running in the node environment */
function isNode() {
	return Boolean(
		typeof process !== 'undefined' && process.versions && process.versions.node
	)
}
exports.isNode = isNode
/** Whether or not stdout and stderr are interactive. */
function isTTY() {
	return (
		isNode() &&
		process.stdout &&
		process.stdout.isTTY === true &&
		process.stderr &&
		process.stderr.isTTY === true
	)
}
exports.isTTY = isTTY
/** Convert a value to its boolean equivalent */
function bool(value) {
	if (value === 'no' || value === 'false' || value === 'n' || value === 'N') {
		return false
	}
	if (value === 'null' || value === 'NULL') return null
	return Boolean(value)
}
exports.bool = bool
/** Whether or not colors are desired on this environment */
function useColors() {
	const env = isNode() && bool(process.env.COLOR || process.env.COLORS)
	if (typeof env === 'boolean') return env
	if (isTTY()) return true
	return (
		isNode() &&
		!process.argv.includes('--no-colors') &&
		!process.argv.includes('--no-color')
	)
}
exports.useColors = useColors
/** Applies the color to the value if desired */
function color(value, color) {
	return useColors() ? color(value) : value
}
exports.color = color
/**
 * Checks to see if a value is an object
 * https://github.com/bevry/typechecker/blob/69008d42927749d7e21cfe9816e478dd8d15ab88/source/index.js#L22-L30
 */
function isObject(value) {
	// null is object, hence the extra check
	return value !== null && typeof value === 'object'
}
/**
 * Return a stringified version of the value with indentation and colors where applicable.
 * Colors will be applied if the environment supports it (--no-colors not present and TTY).
 */
function inspect(value, opts = {}) {
	// If the terminal supports colours, and the user hasn't specified, then default to a sensible default
	const colors = useColors()
	const depth = 50
	// Inspect and return using our defaults
	return util_1.default.inspect(value, { colors, depth, ...opts })
}
exports.inspect = inspect
/** Return the difference between the new data and the old data. */
function diff(newData, oldData) {
	// robust check for object, and try catch
	// to prevent https://github.com/bevry/assert-helpers/issues/5
	if (isObject(newData) && isObject(oldData)) {
		try {
			return diff_1.diffJson(oldData, newData)
		} catch (err) {
			// continue
		}
	}
	// otherwise, continue with classical inspection
	const a = inspect(oldData, { colors: false })
	const b = inspect(newData, { colors: false })
	return diff_1.diffChars(a, b)
}
exports.diff = diff
/** Return a highlighted string of a difference. */
function inspectDiffResult(d) {
	const colors = useColors()
	const result = d.reduce(function (accumulator, part) {
		let value = part.value
		if (colors) {
			if (part.added) {
				value =
					ansicolors_1.default.open.black +
					ansicolors_1.default.bgGreen(value) +
					ansicolors_1.default.open.green
			} else if (part.removed) {
				value =
					ansicolors_1.default.open.black +
					ansicolors_1.default.bgBrightRed(value) +
					ansicolors_1.default.open.green
			}
		}
		return accumulator + value
	}, '')
	return colors ? ansicolors_1.default.green(result) : result
}
exports.inspectDiffResult = inspectDiffResult
/** Return the highlighted comparison between the new data and the old data. */
function compare(newData, oldData) {
	return inspectDiffResult(diff(newData, oldData))
}
exports.compare = compare
/** Alias for {@link compare} */
function inpectDiff(newData, oldData) {
	return compare(newData, oldData)
}
exports.inpectDiff = inpectDiff
/** Log the inspected values of each of the arguments to stdout */
function log(...args) {
	if (isNode() && process.env.ASSERT_SILENCE) return
	for (let i = 0; i < args.length; ++i) {
		console.log(inspect(args[i]))
	}
}
exports.log = log
/** Output a comparison of the failed result to stderr */
function logComparison(actual, expected, error) {
	if (isNode() && process.env.ASSERT_SILENCE) return
	const lines = [
		'------------------------------------',
		'Comparison Error:',
		color(error.stack || error.message || error, ansicolors_1.default.green),
		'',
	]
	lines.push('Comparison Diff:', compare(actual, expected), '')
	lines.push(
		'Comparison Actual:',
		inspect(actual),
		'',
		'Comparison Expected:',
		inspect(expected),
		'------------------------------------'
	)
	// Work for node
	if (isNode() && process.stderr) {
		process.stderr.write(lines.join('\n') + '\n')
	}
	// Work for browsers
	else {
		console.log(lines.join('\n'))
	}
}
exports.logComparison = logComparison
/** Same as assert.strictEquals in that it performs a strict equals check, but if a failure occurs it will output detailed information */
function equal(actual, expected, testName = 'equal assertion', next) {
	try {
		assert_1.default.strictEqual(actual, expected, testName)
	} catch (checkError) {
		logComparison(actual, expected, checkError)
		if (next) {
			next(checkError)
			return
		} else {
			throw checkError
		}
	}
	if (next) next()
}
exports.equal = equal
/** Ensure what is passed is undefined, otherwise fail and output what it is */
function undef(actual, testName = 'undef assertion', next) {
	try {
		assert_1.default.strictEqual(typeof actual, 'undefined', testName)
	} catch (checkError) {
		logComparison(actual, 'undefined', checkError)
		if (next) {
			next(checkError)
			return
		} else {
			throw checkError
		}
	}
	if (next) next()
}
exports.undef = undef
/** Ensure what is passed is undefined or null, otherwise fail and output what it is */
function nullish(actual, testName = 'nullish assertion', next) {
	try {
		assert_1.default.strictEqual(typeof actual, 'undefined', testName)
	} catch (e1) {
		try {
			assert_1.default.strictEqual(actual, null, testName)
		} catch (e2) {
			const error = new errlop_1.default(e2, e1)
			logComparison(actual, 'nullish', error)
			if (next) {
				next(error)
				return
			} else {
				throw error
			}
		}
	}
	if (next) next()
}
exports.nullish = nullish
/** Same as assert.deepStrictEqual in that it performs a deep strict equals check, but if a failure occurs it will output detailed information */
function deepEqual(actual, expected, testName = 'deep equal assertion', next) {
	try {
		assert_1.default.deepStrictEqual(actual, expected, testName)
	} catch (checkError) {
		logComparison(actual, expected, checkError)
		if (next) {
			next(checkError)
			return
		} else {
			throw checkError
		}
	}
	if (next) next()
}
exports.deepEqual = deepEqual
/** Checks to see if the actual result contains the expected result .*/
function contains(actual, expected, testName = 'contains assertion', next) {
	if (testName == null)
		testName = `Expected [${actual}] to contain [${expected}]`
	try {
		assert_1.default.ok(actual.indexOf(expected) !== -1, testName)
	} catch (checkError) {
		if (next) {
			next(checkError)
			return
		} else {
			throw checkError
		}
	}
	if (next) next()
}
exports.contains = contains
/** Checks to see if an error was as expected, if a failure occurs it will output detailed information */
function errorEqual(
	actualError,
	expectedError,
	testName = 'error equal assertion',
	next
) {
	let expectedErrorMessage, actualErrorMessage
	if (expectedError) {
		if (expectedError instanceof Error) {
			expectedErrorMessage = expectedError.message
		} else {
			expectedErrorMessage = expectedError
			expectedError = new Error(expectedErrorMessage)
		}
	}
	if (actualError) {
		if (actualError instanceof Error) {
			actualErrorMessage = actualError.message
		} else {
			actualErrorMessage = actualError
			actualError = new Error(actualErrorMessage)
		}
	}
	try {
		if (actualErrorMessage && expectedErrorMessage) {
			contains(actualErrorMessage, expectedErrorMessage, testName)
		} else {
			equal(actualError, expectedError || null, testName)
		}
	} catch (checkError) {
		logComparison(
			actualError && (actualError.stack || actualError.message || actualError),
			expectedErrorMessage,
			checkError
		)
		if (next) {
			next(checkError)
			return
		} else {
			throw checkError
		}
	}
	if (next) next()
}
exports.errorEqual = errorEqual
/** Generate a callback that will return the specified value. */
function returnViaCallback(value) {
	return function () {
		return value
	}
}
exports.returnViaCallback = returnViaCallback
/** Generate a callback that will receive a completion callback whcih it will call with the specified result after the specified delay. */
/* eslint no-magic-numbers:0 */
function completeViaCallback(value, delay = 100) {
	return function (complete) {
		wait(delay, function () {
			complete(null, value)
		})
	}
}
exports.completeViaCallback = completeViaCallback
/** Generate a callback that will receive a completion callback which it will call with the passed error after the specified delay. */
/* eslint no-magic-numbers:0 */
function errorViaCallback(error, delay = 100) {
	return function (complete) {
		wait(delay, function () {
			complete(error)
		})
	}
}
exports.errorViaCallback = errorViaCallback
/** Generate a callback that return an error instance with the specified message/error. */
function returnErrorViaCallback(error = 'an error occured') {
	return function () {
		if (error instanceof Error) {
			return error
		} else {
			return new Error(error)
		}
	}
}
exports.returnErrorViaCallback = returnErrorViaCallback
/** Generate a callback that throw an error instance with the specified message/error. */
function throwErrorViaCallback(error = 'an error occured') {
	return function () {
		if (error instanceof Error) {
			throw error
		} else {
			throw new Error(error)
		}
	}
}
exports.throwErrorViaCallback = throwErrorViaCallback
/** Generate a callback that will check the arguments it received with the arguments specified, if a failure occurs it will output detailed information. */
function expectViaCallback(...expected) {
	return (...actual) => deepEqual(actual, expected)
}
exports.expectViaCallback = expectViaCallback
/**
 * Generate a callback that will check its error (the actual error) against the passed error (the expected error).
 * If a failure occurs it will output detailed information. */
function expectErrorViaCallback(
	expected,
	testName = 'expect error via callback assertion',
	next
) {
	return (actual) => errorEqual(actual, expected, testName, next)
}
exports.expectErrorViaCallback = expectErrorViaCallback
/** Expect the passed function to throw an error at some point. */
function expectThrowViaFunction(
	expected,
	fn,
	testName = 'expect error via function assertion',
	next
) {
	let actual = null
	try {
		fn()
	} catch (error) {
		actual = error
	}
	errorEqual(actual, expected, testName, next)
}
exports.expectThrowViaFunction = expectThrowViaFunction
/** @deprecated Use {@link expectErrorViaFunction} instead */
function expectErrorViaFunction() {
	throw new Error(
		'expectErrorViaFunction has been deprecated, use expectThrowViaFunction instead'
	)
}
exports.expectErrorViaFunction = expectErrorViaFunction
/** @deprecated Use {@link expectErrorViaFunction} instead */
function expectFunctionToThrow() {
	throw new Error(
		'expectFunctionToThrow has been deprecated, use expectThrowViaFunction instead'
	)
}
exports.expectFunctionToThrow = expectFunctionToThrow
